﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Net;
using System.Net.Sockets;
using System.Reflection;
using System.Text;
using System.Runtime.InteropServices;
using System.Windows.Forms;
using Microsoft.Win32;
using TSTunnels.Client.WtsApi32;
using TSTunnels.Common;

namespace TSTunnels.Client
{
	public class Client
	{
		private const string registryAddins = @"Software\Microsoft\Terminal Server Client\Default\AddIns\";

		[ExportDllAttribute.ExportDll("DllRegisterServer", CallingConvention.StdCall)]
		public static void DllRegisterServer()
		{
			using (var key = Registry.CurrentUser.CreateSubKey(registryAddins + ChannelMessage.ChannelName))
			{
				Debug.Assert(key != null);
				var location = typeof(Client).Assembly.Location;
				Debug.Assert(location != null);
				key.SetValue("Name", location);
			}
		}

		[ExportDllAttribute.ExportDll("DllUnregisterServer", CallingConvention.StdCall)]
		public static void DllUnregisterServer()
		{
			Registry.CurrentUser.DeleteSubKey(registryAddins + ChannelMessage.ChannelName);
		}

		private static Client client;

		[ExportDllAttribute.ExportDll("VirtualChannelEntry", CallingConvention.StdCall)]
		public static bool VirtualChannelEntry(ref ChannelEntryPoints entry)
		{
			client = new Client(entry);
			return client.VirtualChannelInit();
		}

		private readonly ChannelEntryPoints EntryPoints;
		private readonly ChannelInitEventDelegate channelInitEventDelegate;
		private readonly ChannelOpenEventDelegate channelOpenEventDelegate;

		private IntPtr Channel;
		private int OpenChannel;

		private Client(ChannelEntryPoints entry)
		{
			EntryPoints = entry;
			channelInitEventDelegate = VirtualChannelInitEventProc;
			channelOpenEventDelegate = VirtualChannelOpenEvent;
		}

		private const string commonAssemblyName = "TSTunnels.Common, Version=1.0.0.0, Culture=neutral, PublicKeyToken=null";
		static Assembly CurrentDomain_AssemblyResolve(object sender, ResolveEventArgs args)
		{
			if (args.Name == commonAssemblyName)
			{
				var location = typeof(Client).Assembly.Location;
				return Assembly.LoadFrom(Path.Combine(Path.GetDirectoryName(location), "TSTunnels.Common.dll"));
			}
			return null;
		}

		private bool VirtualChannelInit()
		{
			AppDomain.CurrentDomain.AssemblyResolve += CurrentDomain_AssemblyResolve;
			try
			{
				Assembly.Load(commonAssemblyName);
			}
			catch (Exception ex)
			{
				MessageBox.Show("TSTunnels: Failed to load TSTunnels.Common.dll.\n\n" + ex.Message, "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
				return false;
			}
			AppDomain.CurrentDomain.AssemblyResolve -= CurrentDomain_AssemblyResolve;
			var cd = new ChannelDef[1];
			cd[0] = new ChannelDef { name = ChannelMessage.ChannelName };
			var ret = EntryPoints.VirtualChannelInit(ref Channel, cd, 1, 1, channelInitEventDelegate);
			if (ret != ChannelReturnCodes.Ok)
			{
				MessageBox.Show("TSTunnels: RDP Virtual channel Init Failed.\n" + ret, "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
				return false;
			}
			return true;
		}

		private void VirtualChannelInitEventProc(IntPtr initHandle, ChannelEvents Event, byte[] data, int dataLength)
		{
			Debug.WriteLine(DateTime.Now + ": VirtualChannelInitEventProc: " + Event);
			switch (Event)
			{
				case ChannelEvents.Initialized:
					break;
				case ChannelEvents.Connected:
					var ret = EntryPoints.VirtualChannelOpen(initHandle, ref OpenChannel, ChannelMessage.ChannelName, channelOpenEventDelegate);
					if (ret != ChannelReturnCodes.Ok)
					{
						MessageBox.Show("TSTunnels: Open of RDP virtual channel failed.\n" + ret, "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
					}
					else
					{
						/*main = new frmMain(EntryPoints, OpenChannel);
						main.Show();
						main.Hide();
						string servername = System.Text.Encoding.Unicode.GetString(data);
						servername = servername.Substring(0, servername.IndexOf('\0'));
						main.Text = "TS addin in C#: " + servername;*/
					}
					break;
				case ChannelEvents.V1Connected:
					MessageBox.Show("TSTunnels: Connecting to a non Windows 2000 Terminal Server.", "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
					break;
				case ChannelEvents.Disconnected:
					//main.RealClosing = true;
					//main.Close();
					break;
				case ChannelEvents.Terminated:
					break;
			}
		}

		private MemoryStream memoryStream;
		public void VirtualChannelOpenEvent(int openHandle, ChannelEvents Event, IntPtr pData, int dataLength, uint totalLength, ChannelFlags dataFlags)
		{
			Debug.WriteLine(DateTime.Now + ": VirtualChannelOpenEvent: " + Event);
			switch (Event)
			{
				case ChannelEvents.DataReceived:
					var data = new byte[dataLength];
					Marshal.Copy(pData, data, 0, dataLength);
					switch (dataFlags & ChannelFlags.Only)
					{
						case ChannelFlags.Only:
							MessageReceived(ChannelMessage.FromByteArray(data));
							break;
						case ChannelFlags.First:
							memoryStream = new MemoryStream((int)totalLength);
							memoryStream.Write(data, 0, data.Length);
							break;
						case ChannelFlags.Middle:
							if (memoryStream != null)
							{
								memoryStream.Write(data, 0, data.Length);
							}
							break;
						case ChannelFlags.Last:
							if (memoryStream != null)
							{
								memoryStream.Write(data, 0, data.Length);
								memoryStream.Position = 0;
								MessageReceived(ChannelMessage.FromStream(memoryStream));
								memoryStream = null;
							}
							break;
					}
					break;
				case ChannelEvents.WriteCanceled:
				case ChannelEvents.WriteComplete:
					/*
					 * The VirtualChannelWrite function is asynchronous. When the write operation has been completed,
					 * your VirtualChannelOpenEvent function receives a CHANNEL_EVENT_WRITE_COMPLETE notification.
					 * Until that notification is received, the caller must not free or reuse the pData buffer passed to VirtualChannelWrite
					 */
					Marshal.FreeHGlobal(pData);
					break;
			}
		}

		private void WriteMessage(ChannelMessage msg)
		{
			var data = msg.ToByteArray();
			var len = 4 + data.Length;
			var ptr = Marshal.AllocHGlobal(len);
			Marshal.WriteInt32(ptr, 0, data.Length);
			Marshal.Copy(data, 0, new IntPtr(ptr.ToInt32() + 4), data.Length);
			var ret = EntryPoints.VirtualChannelWrite(OpenChannel, ptr, (uint)len, ptr);
		}

		private readonly Dictionary<int, Stream> streams = new Dictionary<int, Stream>();
		private int connectionCount;

		public void MessageReceived(ChannelMessage msg)
		{
			switch (msg.Type)
			{
				case MessageType.HelloRequest:
					{
						WriteMessage(new HelloReply());
					}
					break;
				case MessageType.ConnectRequest:
					{
						var request = (ConnectRequest)msg;
						var tcpClient = new TcpClient();
						tcpClient.BeginConnect(request.Address, request.Port, client_Connect, new KeyValuePair<int, TcpClient>(request.StreamIndex, tcpClient));
					}
					break;
				case MessageType.ListenRequest:
					{
						var request = (ListenRequest)msg;
						var listener = new TcpListener(IPAddress.Parse(request.Address), request.Port);
						listener.Start();
						listener.BeginAcceptTcpClient(listener_AcceptTcpClient, (AcceptTcpClientCallback)delegate(IAsyncResult ar)
						{
							var streamIndex = --connectionCount;
							var stream = streams[streamIndex] = listener.EndAcceptTcpClient(ar).GetStream();
							WriteMessage(new ListenResponse(request.StreamIndex, streamIndex));
							new StreamPump(stream,
								data => WriteMessage(new StreamData(streamIndex, data)),
								ex => WriteMessage(new StreamError(streamIndex, ex.ToString()))).Pump();
							return listener;
						});
					}
					break;
				case MessageType.StreamData:
					{
						var data = (StreamData)msg;
						var stream = streams[data.StreamIndex];
						try
						{
							stream.Write(data.Data, 0, data.Data.Length);
						}
						catch (Exception ex)
						{
							WriteMessage(new StreamError(data.StreamIndex, ex.ToString()));
						}
					}
					break;
				case MessageType.StreamError:
					{
						var error = (StreamError)msg;
						var stream = streams[error.StreamIndex];
						if (error.Exception.Length != 0)
						{
							try
							{
								var data = Encoding.UTF8.GetBytes(error.Exception);
								stream.Write(data, 0, data.Length);
							}
							catch (IOException)
							{
							}
							return;
						}
						stream.Close();
						streams.Remove(error.StreamIndex);
					}
					break;
			}
		}

		private delegate TcpListener AcceptTcpClientCallback(IAsyncResult ar);
		private static void listener_AcceptTcpClient(IAsyncResult ar)
		{
			var callback = (AcceptTcpClientCallback)ar.AsyncState;
			var listener = callback(ar);
			listener.BeginAcceptTcpClient(listener_AcceptTcpClient, callback);
		}

		private void client_Connect(IAsyncResult ar)
		{
			var state = (KeyValuePair<int, TcpClient>)ar.AsyncState;
			var streamIndex = state.Key;
			var tcpClient = state.Value;
			try
			{
				tcpClient.EndConnect(ar);
				var stream = streams[streamIndex] = tcpClient.GetStream();
				new StreamPump(stream,
					data => WriteMessage(new StreamData(streamIndex, data)),
					ex => WriteMessage(new StreamError(streamIndex, ex.ToString()))).Pump();
				WriteMessage(new ConnectResult(streamIndex, tcpClient.Client.RemoteEndPoint.ToString()));
			}
			catch (Exception ex)
			{
				WriteMessage(new StreamError(streamIndex, ex.ToString()));
				Debug.Print(ex.ToString());
			}
		}
	}
}
